/* ****************************************************************
 * Copyright Stanford University 1998,99 - All Rights Reserved
 ****************************************************************** 
 
 * Permission to use, copy, modify, and distribute this software 
 * and its documentation for any purpose is hereby granted without 
 * fee, provided that the above copyright notice appears in all copies
 * and that both the copyright notice, this permission notice, and 
 * the following disclaimer appear in supporting documentation, and 
 * that the name of Stanford University, not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * STANFORD UNIVERSITY, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS. IN NO EVENT SHALL STANFORD UNIVERSITY BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION 
 * OF CONTRACT, NEGLIGENCE OR OTHER ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


/*
 * The Sim Web Site : http://klamath.stanford.edu/tools/SIM
 * The SIM Mailing List: sim-simulator@lists.stanford.edu
 
 * Send mail to the above email address with "subscribe sim-simulator" in
 * the body of the message.
 *
 */

#include <string.h>
#include <assert.h>
#include "sim.h"
#include "algorithm.h"
#include "rr.h"
#include "scheduleStats.h"
#include "miscfns.h"


static int selectGrant();
static int selectAccept();
static void createScheduleState();
static InputSchedulerState *createInScheduler();
static OutputSchedulerState *createOutScheduler();
static int l_search() ;
static void dtc_sort() ;
static void setupGrants() ;
static void setupAccepts() ;
static void displayGrantTokens() ;
static void displayCreditTokens() ;
static int calculateGrantTokens() ;
static void calculateCreditTokens() ;
static int scheduleCreditTokens() ;
static int remainingGrantTokens() ;
static void resetGrantTokens() ;
static void resetRemainingGrantTokens() ;

static void calculateGrants() ;
static void displayGrants() ;
static int scheduleGrants() ;


typedef struct {	
	struct Element *next;
	struct Element *prev;
	int *inputAcceptPointer;
	int *outputGrantPointer;
	int *connection;
} State;

void
pbcGBVOQ(action, aSwitch, argc, argv)
SwitchAction action;
Switch *aSwitch;
int argc;
char **argv;
{
	SchedulerState 			*scheduleState;
	InputSchedulerState 	*inputSchedule;
	OutputSchedulerState	*outputSchedule;
	int input, switchOutput, fabricOutput, numFabricOutputs, credits, i, j ;
	Cell *aCell;
	struct List *fifo;
	
	if(debug_algorithm)
		printf("Algorithm 'pbcGBVOQ()' called by switch %d\n", aSwitch->switchNumber);
	
	switch(action) {
		case SCHEDULING_USAGE:
			fprintf(stderr, "Options for \"pbcGBVOQ\" scheduling algorithm:\n");
			fprintf(stderr, "    -n number_of_iterations. Default (is overrided): 1\n");
			break;
		case SCHEDULING_INIT:	
			if(debug_algorithm) printf("	SCHEDULING_INIT\n");
			
			if( aSwitch->scheduler.schedulingState == NULL )
			{
				extern int opterr;
				extern int optopt;
				extern int optind;
				extern char *optarg;
				int c;
				
				
				/* Allocate scheduler state for this switch */
				createScheduleState(aSwitch);
				scheduleState = (SchedulerState *) aSwitch->scheduler.schedulingState;
				scheduleState->numIterations=1;
				
				opterr=0;
				optind=1;
				while( (c = getopt(argc, argv, "n:")) != EOF)
					switch (c)
				{
					case 'n':
						scheduleState->numIterations = 1 ;	/* Override : In pbcGBVOQ scheduling, there is no definition about the algorithm s iterations. */
						break;
					case '?':
						fprintf(stderr, "-------------------------------\n");
						fprintf(stderr, "pbcGBVOQ: Unrecognized option -%c\n", optopt);
						pbcGBVOQ(SCHEDULING_USAGE);
						fprintf(stderr, "--------------------------------\n");
						exit(1);
					default:
						break;
				}
				
				
				/* Reset and initialize scheduling stats for this switch. */
				scheduleCellStats( SCHEDULE_CELL_STATS_INIT, aSwitch );
				
				printf("numIterations %d\n", scheduleState->numIterations);
			}
			
			break;
			
		case SCHEDULING_EXEC:
		{
			int iteration;
			int numConnectionsThisIteration;
			
			if(debug_algorithm) printf("	SCHEDULING_EXEC\n");
			
			scheduleState=(SchedulerState *)aSwitch->scheduler.schedulingState;
			numFabricOutputs = aSwitch->numOutputs * 
			aSwitch->fabric.Xbar_numOutputLines;
			
			/*************** INITIALIZE ***************/
			/* Set up inputs (grants and accepts) */
			for(input=0; input<aSwitch->numInputs; input++)
			{
				inputSchedule = scheduleState->inputSched[input];
				inputSchedule->accept = NONE;
				inputSchedule->gbvoq_accept = NONE ;
				
				for(fabricOutput=0;fabricOutput<numFabricOutputs;fabricOutput++)
					inputSchedule->grant[fabricOutput] = 0;
			}
			
			/* Set up outputs (accepts) */
			for(fabricOutput=0;fabricOutput<numFabricOutputs;fabricOutput++)
			{
				outputSchedule = scheduleState->outputSched[fabricOutput];
				outputSchedule->accept=NONE;
				outputSchedule->grant=NONE;
			}
			
			if( debug_algorithm ) fprintf(stdout, "$$$ PBC $$$ : A | ... R E S E T \n") ;
			/* Reset grants for all inputs : GQ(i,*) = NONE . */
			for(input=0; input<aSwitch->numInputs; input++)
			{
				resetGrantTokens(aSwitch, input) ;
				if( debug_algorithm ) displayGrantTokens(aSwitch, input) ;
			}
			/* Calculate credits for all outputs : CQ(j) = ? . */
			for(fabricOutput=0;fabricOutput<numFabricOutputs;fabricOutput++)
				calculateCreditTokens(aSwitch, fabricOutput) ;
			if( debug_algorithm ) 
			{
				displayCreditTokens(aSwitch) ;
				fprintf(stdout, "$$$ PBC $$$ : free tokens | %d \n", scheduleCreditTokens(aSwitch)) ;
			}
			
			/* Check for newly arrived cells at head of input queues. */
			for(input=0; input<aSwitch->numInputs; input++)
				for(switchOutput=0; switchOutput<aSwitch->numOutputs; 
					switchOutput++)
				{
					fifo = aSwitch->inputBuffer[input]->fifo[switchOutput];
					if( fifo->number )
					{
						aCell = fifo->head->Object;
						scheduleCellStats(SCHEDULE_CELL_STATS_HEAD_ARRIVAL,
										  aSwitch, aCell);
					}
				}
			
			/* Calculate and update synchronization stats */
			countSynch(aSwitch);
			/***************  END INITIALIZE ***************/
			
			
			/*****************************/
			/*****   Main Loop  **********/
			/*****************************/
			for(iteration=0;iteration<scheduleState->numIterations;iteration++)
			{
				numConnectionsThisIteration = 0;
				if( debug_algorithm )	fprintf(stdout, "---> TIME : %ld \n", now) ;
				
				for(fabricOutput=0;fabricOutput<numFabricOutputs;fabricOutput++)
				{
					/* NOTE : The grant pointers are always incremented by 1, irrespective of the accept/drop outcome. */
					scheduleState->outputSched[fabricOutput]->unsynch = scheduleState->outputSched[fabricOutput]->last_accepted_grant ;
					
					setupGrants(aSwitch, fabricOutput) ;
					calculateGrants(aSwitch, fabricOutput) ;
					
					while( aSwitch->outputBuffer[fabricOutput]->gbvoqQueue > 0 )
					{
						credits = aSwitch->outputBuffer[fabricOutput]->gbvoqQueue ;
						/* grant an input for this output line */
						if(debug_algorithm)	printf("Selecting grant for output %d\n", fabricOutput);
						
						input = selectGrant(aSwitch, fabricOutput);
						if( debug_algorithm )	fprintf(stdout, "*** 4 *** : [out]=%d was GRANTED to [in]=%d *** 4 *** \n", fabricOutput, input) ;
						if( debug_algorithm )	fprintf(stdout, "---> Main Loop | ITERATION=%d | OUT:%d \t calculated_input : %d \n", iteration, fabricOutput, input) ;
						
						if(debug_algorithm)	printf("Selected input %d\n", input);
						if( input != NONE ) 
						{
							inputSchedule = scheduleState->inputSched[input];
							outputSchedule=scheduleState->outputSched[fabricOutput];
							/* The output notifies each input whether or not its request was granted. */
							inputSchedule->grant[fabricOutput] = 1;
							if( debug_algorithm )	fprintf(stdout, "*** 5 *** : inputSchedule(%d)->grant[%d] = %d *** 5 *** \n", input, fabricOutput, inputSchedule->grant[fabricOutput]) ;
							outputSchedule->grant = input;
							if( debug_algorithm )	fprintf(stdout, "*** 6 *** : outputSchedule(%d)->grant = input(%d) *** 6 *** \n", fabricOutput, input) ;
							
							aSwitch->inputBuffer[input]->grantQueue[fabricOutput] = 1 ;
							aSwitch->outputBuffer[fabricOutput]->gbvoqQueue -- ;
						}
						if( credits == aSwitch->outputBuffer[fabricOutput]->gbvoqQueue )	break ;	/* There are no available credits anymore for this output. */
					}
					/* NOTE : The grant pointers are always incremented by 1, irrespective of the accept/drop outcome. */
					scheduleState->outputSched[fabricOutput]->last_accepted_grant = (scheduleState->outputSched[fabricOutput]->unsynch + 1)%(aSwitch->numInputs) ;
				}
				
				if( debug_algorithm ) 
				{
					fprintf(stdout, "$$$ PBC $$$ : B | ... U P D A T E \n") ;
					displayGrants(aSwitch) ;
					fprintf(stdout, "$$$ PBC $$$ : gbvoq tokens | %d \n", scheduleGrants(aSwitch)) ;
					for(i=0; i<aSwitch->numInputs; i++)
						displayGrantTokens(aSwitch, i) ;
				}
					
				for(input=0; input<aSwitch->numInputs; input++)
				{
					/* accept output for each input  */
					setupAccepts(aSwitch, input) ;
					
					fabricOutput = selectAccept(aSwitch, input);
					if( debug_algorithm )	fprintf(stdout, "*** 10 *** : [in]=%d was ACCEPTED by [out]=%d *** 10 *** \n", input, fabricOutput) ;
					if( debug_algorithm )	fprintf(stdout, "---> Main Loop | ITERATION=%d | IN:%d \t calculated_fabricOutput : %d \n", iteration, input, fabricOutput) ;
					
					if( fabricOutput >= 0 && (aSwitch->inputBuffer[input]->grantQueue[fabricOutput] == 1) ) 
					{
						// fprintf(stderr, "NOTE : Main Loop : %ld | (%d,%d) | %d \n", now, input, fabricOutput, aSwitch->outputBuffer[fabricOutput]->creditQueue) ;
						numConnectionsThisIteration++;
						inputSchedule = scheduleState->inputSched[input];
						outputSchedule=scheduleState->outputSched[fabricOutput];
						inputSchedule->accept = fabricOutput;
						if( debug_algorithm )	fprintf(stdout, "*** 11 *** : inputSchedule(%d)->accept = fabricOutput(%d) *** 11 *** \n", input, fabricOutput) ;
						outputSchedule->accept = input;
						if( debug_algorithm )	fprintf(stdout, "*** 12 *** : outputSchedule(%d)->accept = input(%d) *** 12 *** \n", fabricOutput, input) ;
						
						inputSchedule->last_accept = (inputSchedule->last_accept + 1)%(aSwitch->numInputs) ;
						/* The status of the CQ is updated during output scheduling (@subportOutputAction). */
						
						/* ISLIP: Accepted, so update grant schedule */
						/* ISlip: move grant scheduler iff first iteration */
						/* Bug fix done - Sundar. */
						// if( iteration == 0 )	outputSchedule->last_accepted_grant = input; 
					}	
				}
				
				if( debug_algorithm ) 
				{
					fprintf(stdout, "$$$ PBC $$$ : C | ... U P D A T E (cont.) \n") ;
					displayGrants(aSwitch) ;
					fprintf(stdout, "$$$ PBC $$$ : gbvoq tokens | %d \n", scheduleGrants(aSwitch)) ;
					for(i=0; i<aSwitch->numInputs; i++)
						displayGrantTokens(aSwitch, i) ;
				}
				
				/* If no connections were made during this iteration, */
				/* the algorithm has converged and so there is no point */
				/* in continuing. */
				if (!numConnectionsThisIteration)
					break;
			}
			scheduleStats(SCHEDULE_STATS_NUM_ITERATIONS, aSwitch, &iteration);
			
			if( debug_algorithm )
			{
				fprintf(stdout, "$$$ PBC $$$ : D | ... T R A N S M I T \n") ;
				for(i=0; i<aSwitch->numInputs; i++)
					for(j=0; j<aSwitch->numOutputs; j++)
						if( aSwitch->inputBuffer[i]->grantQueue[j] == 1 )	fprintf(stdout, "GQ(%d,%d) ", i, j) ;
				fprintf(stdout, "\n") ;
			}
			
			for(fabricOutput=0; fabricOutput<numFabricOutputs; fabricOutput++)
			{
				input = scheduleState->outputSched[fabricOutput]->accept;
				if( debug_algorithm )	fprintf(stdout, "*** 13 *** : input(%d) = scheduleState->outputSched[fabricOutput(%d)]->accept *** 13 *** \n", input, fabricOutput) ;
				/* NOTE : The "aSwitch->fabric.Xbar_matrix" data structure is overriden by writting to the output buffers (crossbar.c) right away. */
				/* This is done for 2 reasons : (a) the schedulers work on a per fabricOutput basis and not in a per switchOutput basis (thus, there is no need for modification of the schedulers) */
				/*                              (b) i sould not mess with the traffic schemes which provide additional FIFOs per input (the schedulers are depedent on this), whereas the PBC does not create any additional FIFOs, so the "numFabricOutputs" changes and adds complexity.  */
				aSwitch->fabric.Xbar_matrix[fabricOutput].input = input;
				if(input!=NONE)	
					aSwitch->fabric.Xbar_matrix[fabricOutput].cell = (Cell *)aSwitch->inputBuffer[input]->fifo[fabricOutput]->head->Object;
			}
			
			break;
		}
			
		case SCHEDULING_REPORT_STATS:
			scheduleStats( SCHEDULE_STATS_PRINT_ALL, aSwitch );
			scheduleCellStats( SCHEDULE_CELL_STATS_PRINT_ALL, aSwitch );
			break;
		case SCHEDULING_INIT_STATS:
			scheduleStats( SCHEDULE_STATS_INIT, aSwitch );
			scheduleCellStats( SCHEDULE_CELL_STATS_INIT, aSwitch );
			break;
		case SCHEDULING_REPORT_STATE:
			scheduleState=(SchedulerState *) aSwitch->scheduler.schedulingState;
			
			numFabricOutputs = aSwitch->numOutputs * 
			aSwitch->fabric.Xbar_numOutputLines;
			
			printf("Output: ");
			for(fabricOutput=0; fabricOutput<numFabricOutputs; fabricOutput++)
				printf("%2d ", fabricOutput);
			printf("Input: ");
			for(input=0; input<aSwitch->numInputs; input++)
				printf("%2d ", input);
			printf("\n        ");
			for(fabricOutput=0; fabricOutput<numFabricOutputs; fabricOutput++)
				printf("%2d ", 
					   ((scheduleState->outputSched[fabricOutput]->last_accepted_grant) + 1)%aSwitch->numInputs);
			printf("       ");
			for(input=0; input<aSwitch->numInputs; input++)
				printf("%2d ", 
					   ((scheduleState->inputSched[input]->last_accept) 
						+ 1)%numFabricOutputs);
			printf("\n");
			break;
		case SCHEDULING_CHECK_STATE_PERIOD:	
		{
			
#ifdef ERRORS_CHECKED
			/* This portion is not included as it has bugs yet to be
			 corrected */
			
			/* State is: state of schedulers and connection matrix. */
			/* Search visitedStateList from head to see if already
			 visited this state. 
			 If so, print from that point to end of list. This is one period.  
			 Else, add state to visitedStateList 
			 */
			int period, foundMatch;
			State *aState;
			State *currentState=(State *) malloc(sizeof(State));
			static struct List visitedStateList=NULL;
			struct Element *anElement;
			int **numConnections;
			int totalNumConnections=0;
			
			if(visitedStateList == NULL)
				visitedStateList = createList();
			numFabricOutputs = aSwitch->numOutputs * 
			aSwitch->fabric.Xbar_numOutputLines;
			
			currentState->inputAcceptPointer = 
			(int *) malloc(sizeof(int) * aSwitch->numInputs);
			currentState->outputGrantPointer = 
			(int *) malloc(sizeof(int) * numFabricOutputs);
			currentState->connection = 
			(int *) malloc(sizeof(int) * numFabricOutputs);
			scheduleState=(SchedulerState *) aSwitch->scheduler.schedulingState;
			
			for(input=0; input<aSwitch->numInputs; input++)
			{
				currentState->inputAcceptPointer[input] = ((scheduleState->inputSched[input]->last_accept)+1) % numFabricOutputs;
			}
			for(fabricOutput=0; fabricOutput<numFabricOutputs; fabricOutput++)
			{
				currentState->outputGrantPointer[fabricOutput] = ((scheduleState->outputSched[fabricOutput]->last_accepted_grant)+1) % aSwitch->numInputs;
				currentState->connection[fabricOutput] = aSwitch->fabric.Xbar_matrix[fabricOutput].input;
			}
			
			aState = (State *) visitedStateList.head;
			foundMatch=0; period=0;
			while( aState )
			{
				/* Compare */
				if( foundMatch || ( ( memcmp(aState->inputAcceptPointer, 
											 currentState->inputAcceptPointer,
											 sizeof(int)*aSwitch->numInputs) == 0 ) &&
								   ( memcmp(aState->outputGrantPointer, 
											currentState->outputGrantPointer,
											sizeof(int)*numFabricOutputs) == 0 ) &&
								   ( memcmp(aState->connection, 
											currentState->connection,
											sizeof(int)*numFabricOutputs) == 0 ) ) )
					
				{
					if( !foundMatch )
					{
						/* Visited this state before */
						foundMatch=1;
						
						/* Allocate and init space for accumulatng stats */
						numConnections = (int **) 
						malloc( sizeof(int *) * aSwitch->numInputs);
						for(input=0; input<aSwitch->numInputs; input++)
							numConnections[input] = (int *) 
							malloc( sizeof(int) * numFabricOutputs);
						
						for(input=0; input<aSwitch->numInputs; input++)
							for(fabricOutput=0; fabricOutput<numFabricOutputs; 
								fabricOutput++)
								numConnections[input][fabricOutput]=0;
						
						printf("\n\nCYCLE RESULTS:\n");
						printf("------------------\n");
					}
					
					period++;
					printState(aSwitch, aState);
					
					/* Calculate number of connections for each i-o pair */
					for(fabricOutput=0; fabricOutput<numFabricOutputs; 
						fabricOutput++)
					{
						input = aState->connection[fabricOutput];
						if( input != NONE )
						{
							numConnections[input][fabricOutput]++;
							totalNumConnections++;
						}
						/* ELSE | NOTE : WTF ???. */
					}
					
				}
				aState = (State *) aState->next;
			}
			if( foundMatch )
			{
				printf("Cycle length: %d\n", period);
				printf("Number of connections per cycle: %d\n", totalNumConnections);
				printf("     ");
				for(fabricOutput=0; fabricOutput<numFabricOutputs; 
					fabricOutput++)
					printf("%2d ", fabricOutput);
				printf("\n---------------------\n", fabricOutput);
				for(input=0; input<aSwitch->numInputs; input++)
				{
					printf("%2d | ", input);
					for(fabricOutput=0; fabricOutput<numFabricOutputs; 
						fabricOutput++)
						printf("%2d ", numConnections[input][fabricOutput]);
					printf("\n");
				}
				printf("\n");
				
				exit(0);
			}
			else
			{
				/* Add state to vistedStateList */
				addElement( &visitedStateList, currentState);
			}
#endif
			
			break;
		}
			
		default:
			break;
	}
	
	
	
	if(debug_algorithm)
		printf("Algorithm 'pbcGBVOQ()' completed for switch %d\n", aSwitch->switchNumber);
}

/***********************************************************************/
static int
selectGrant(aSwitch, output)
Switch *aSwitch;
int output;
{
	/*
	 Select next requesting input from schedule
	 */
	int input, in ;
	SchedulerState	*scheduleState =
    (SchedulerState *)aSwitch->scheduler.schedulingState;
	OutputSchedulerState *outputSchedule=scheduleState->outputSched[output];
	InputBuffer  *inputBuffer;
	Cell *aCell, *pipelineCell ;
	
	int increment;
	
	int switchOutput=output/aSwitch->fabric.Xbar_numOutputLines;
	
	
	/* Check to see if output has already been accepted by an input from
	 an earlier iteration */
	if( (outputSchedule->accept != NONE) && !(aSwitch->outputBuffer[output]->gbvoqQueue>0) )
    {
		if(debug_algorithm)
			printf("Output %d already accepted by input %d\n", output, outputSchedule->accept);
		return(NONE);
    }
	
	/**************************************/
	
	for(increment=0; increment<aSwitch->numInputs; increment++)
    {
		/* If an unmatched output receives multiple requests, it chooses the one that appears next
		 ------
		 in a fixed, round-robin schedule starting from the highest priority element.
		 NOTE : GBVOQ strategy indicates that the input at the head of the OPL should be served.
		 ---                   */
		
		if( aSwitch->outputBuffer[output]->outputPriorityList->head != NULL )
		{
			for(in=0; in<aSwitch->numOutputs; in++)
			{
				input = outputSchedule->GBVOQ_grants[in] ;
				if( debug_algorithm )	fprintf(stdout, "*** 1 *** : input(%d) = outputSchedule(%d)->GBVOQ_grants[%d] *** 1 *** \n", input, output, in) ;
				
				/* ******************************************** */
				/* NOTE : PBC functions in a pipelined fashion. */
				/* ******************************************** */
				if(aSwitch->numInternalBuffers != NONE)
					if( input != NONE )
						if( aSwitch->inputBuffer[input]->fifo[switchOutput]->head != NULL )
						{
							pipelineCell = (Cell *)aSwitch->inputBuffer[input]->fifo[switchOutput]->head->Object ;
							if( pipelineCell->commonStats.pipelined == 0 )	{ pipelineCell->commonStats.pipelined = 1 ; continue ; }
						}
				/* ******************************************** */
				
				if( input != NONE )
				{ if( aSwitch->inputBuffer[input]->grantQueue[switchOutput] == NONE/*scheduleState->inputSched[input]->gbvoq_accept == NONE*/ )	break ; }
				else	break ;
			}
			if( (in == aSwitch->numOutputs) && (aSwitch->inputBuffer[input]->grantQueue[switchOutput] != NONE/*scheduleState->inputSched[input]->gbvoq_accept != NONE*/) )
				return NONE ;	// NOTE : if the loop does not break, it should return NONE.
			if( debug_algorithm )	fprintf(stdout, "---> selectGrant() : OUT:%d \t GBVOQ:%d \n", output, input) ;
		}
		else	return NONE ;
		
		if( input == NONE )	return NONE ;	// NOTE : 'else break'
		
		
		inputBuffer = aSwitch->inputBuffer[input];
		if( debug_algorithm )	fprintf(stdout, "*** 2 *** : scheduleState->inputSched[%d]->accept *** 2 *** \n", input, scheduleState->inputSched[input]->accept) ;
		
		if( (inputBuffer->fifo[switchOutput]->number > 0) && (aSwitch->inputBuffer[input]->grantQueue[switchOutput] == NONE/*scheduleState->inputSched[input]->accept == NONE*/) )
		{
			if( debug_algorithm )	fprintf(stdout, "*** 3 *** : scheduleState->inputSched[%d]->accept *** 3 *** \n", input, scheduleState->inputSched[input]->accept) ;
			scheduleState->inputSched[input]->gbvoq_accept = input ;
			/* STATS */
			/* Mark grantTime for cell at head of line */
			aCell = (Cell *) inputBuffer->fifo[switchOutput]->head->Object;
			scheduleCellStats( SCHEDULE_CELL_STATS_UPDATE_GRANT, 
							  aSwitch, aCell);
			/* END STATS */
			
			/* ISLIP: updated only on accept */
			/* outputSchedule->last_accepted_grant = input; */
			if(debug_algorithm)
				printf("Output %d granting to input %d\n", output, input);
			return(input);
		}
    }
	
	if(debug_algorithm)
		printf("No inputs requested output %d. None granted.\n", output);
	
	return(NONE);
	
}

static int
selectAccept(aSwitch, input)
Switch *aSwitch;
int input;
{
	/*
	 Choose output to accept from schedule
	 */
	SchedulerState	*scheduleState=(SchedulerState *)aSwitch->scheduler.schedulingState;
	InputSchedulerState *inputSchedule=scheduleState->inputSched[input];
	
	Cell *aCell ;
	
	int output, switchOutput, numFabricOutputs, out;
	
	int increment;
	
	
	numFabricOutputs = aSwitch->numOutputs*aSwitch->fabric.Xbar_numOutputLines;
	
	if( ! (remainingGrantTokens(aSwitch, input) > 0) )
		return( NONE );
	
	/*	How many outputs granted to this input? */
	/* 
	 numGrants=0;
	 for(output=0; output<numFabricOutputs; output++)
	 {	
	 if( inputSchedule->grant[output] == 1 )
	 numGrants++;
	 }
	 if( numGrants > 1 )
	 printf("%d outputs granted to input %d\n", numGrants, input);
	 */
	
	for( increment=0; increment<numFabricOutputs; increment++ )
    {
		/* If an input receives multiple grants, it chooses the one that appears next
		 -----
		 in a fixed, round-robin schedule starting from the highest priority element.
		 NOTE : GBVOQ strategy indicates that the input at the head of the IPL should be served.
		 ---                   */
		
		if( aSwitch->inputBuffer[input]->inputPriorityList->head != NULL )
		{
			for(out=0; out<aSwitch->numInputs; out++)
			{
				output = inputSchedule->GBVOQ_accepts[out] ;
				if( debug_algorithm )	fprintf(stdout, "*** 7 *** : output(%d)= inputSchedule(%d)->GBVOQ_accepts[%d] *** 7 *** \n", output, input, out) ;
				if( output != NONE )
				{
					if( ! (aSwitch->outputBuffer[output]->creditQueue > 0) )	{ aSwitch->inputBuffer[input]->grantQueue[output] = 0 ; if(debug_algorithm) fprintf(stdout, "pbcGBVOQ : Output %d has no available credits. \n", output) ; continue ; }
					if( aSwitch->inputBuffer[input]->grantQueue[output] == 1/*inputSchedule->grant[output] == 1*/ )	break ; 
				}
				else	break ;
			}
			if( (out == aSwitch->numInputs) && (aSwitch->inputBuffer[input]->grantQueue[output] != 1)/*(inputSchedule->grant[output] != 1)*/ )
				return NONE ;	// NOTE : if the loop does not break, it should return NONE.
			if( debug_algorithm )	fprintf(stdout, "---> selectAccept() | input:%d \t GBVOQ:%d \n", input, output) ;
		}
		else	return NONE ;
		
		if( output == NONE )	return NONE ;	// NOTE : 'else break'
		
		if( debug_algorithm )	fprintf(stdout, "*** 8 *** : inputSchedule(%d)->grant[output(%d)] = %d *** 8 *** \n", input, output, inputSchedule->grant[output]) ;
		if( aSwitch->inputBuffer[input]->grantQueue[output] == 1 /*inputSchedule->grant[output] == 1*/)
		{
			// fprintf(stderr, "NOTE : selectAccept() : %ld | (%d,%d) | %d \n", now, input, output, aSwitch->outputBuffer[output]->creditQueue) ;
			if( debug_algorithm )	fprintf(stdout, "*** 9 *** : inputSchedule(%d)->grant[output(%d)] = %d *** 9 *** \n", input, output, inputSchedule->grant[output]) ;

			// inputSchedule->last_accept = output;
			resetRemainingGrantTokens(aSwitch, input, output) ;
			aSwitch->outputBuffer[output]->creditQueue -- ;
			
			/* STATS */
			/* Mark cell at head of line as accepted */
			switchOutput = output / aSwitch->fabric.Xbar_numOutputLines;
			aCell = (Cell *)
			aSwitch->inputBuffer[input]->fifo[switchOutput]->head->Object;
			scheduleCellStats( SCHEDULE_CELL_STATS_UPDATE_ACCEPT, 
							  aSwitch, aCell);
			
			if(debug_algorithm)
				printf("Input %d accepting output %d\n", input, output);
			return(output);	
		}
    }
	
	if(debug_algorithm)
		printf("Input %d not accepting any output\n", input);
	
	return(NONE);
	
}


/* Create schedule state variables for aSwitch */
static void
createScheduleState(aSwitch)
Switch *aSwitch;
{
	SchedulerState *scheduleState;
	
	
	int input, output, numFabricOutputs;
	
	numFabricOutputs = aSwitch->numOutputs*aSwitch->fabric.Xbar_numOutputLines;
	
	scheduleState = (SchedulerState *) malloc(sizeof(SchedulerState));
	aSwitch->scheduler.schedulingState = scheduleState;
	
	
	scheduleState->inputSched = (InputSchedulerState **) 
    malloc( aSwitch->numInputs * sizeof(InputSchedulerState *) );
	for(input=0; input<aSwitch->numInputs; input++)
		scheduleState->inputSched[input] = createInScheduler(aSwitch, input);
	
	scheduleState->outputSched = (OutputSchedulerState **) 
    malloc( numFabricOutputs * sizeof(OutputSchedulerState *) );
	for(output=0; output<numFabricOutputs; output++)
		scheduleState->outputSched[output]=createOutScheduler(aSwitch, output);
	
}

static InputSchedulerState *
createInScheduler(aSwitch, input)
Switch *aSwitch;
int input;
{	
	SchedulerState 		*scheduleState=aSwitch->scheduler.schedulingState;
	InputSchedulerState	**inputSchedule=scheduleState->inputSched;
	int i, numFabricOutputs;
	
	numFabricOutputs = aSwitch->numOutputs*aSwitch->fabric.Xbar_numOutputLines;
	
	inputSchedule[input] = (InputSchedulerState *) malloc( sizeof( InputSchedulerState ) );
	inputSchedule[input]->grant = (int *) malloc( sizeof(int)*numFabricOutputs);
	
	
	inputSchedule[input]->GBVOQ_accepts = (int *)malloc( sizeof(int)*numFabricOutputs ) ;
	
#ifdef	DTC
	inputSchedule[input]->slackness_accepts = (int *)malloc( sizeof(int)*numFabricOutputs ) ;
	inputSchedule[input]->slackness_outputs = (int *)malloc( sizeof(int)*numFabricOutputs ) ;
#endif
	
	inputSchedule[input]->RR_accepts = (int *) malloc( sizeof(int)*numFabricOutputs);
	for(i=0; i<numFabricOutputs; i++ )
	{
		inputSchedule[input]->RR_accepts[i] = i;
		
		inputSchedule[input]->GBVOQ_accepts[i] = NONE ;
		if( DTC )		
		{
			inputSchedule[input]->slackness_accepts[i] = 123456789 ;
			inputSchedule[input]->slackness_outputs[i] = NONE ;
		}
	}
	inputSchedule[input]->last_accept=0;
	inputSchedule[input]->gbvoq_accept = NONE ;
	
	return(inputSchedule[input]);
	
}

static OutputSchedulerState *
createOutScheduler(aSwitch, output)
Switch *aSwitch;
int output;
{
	SchedulerState 		*scheduleState=aSwitch->scheduler.schedulingState;
	OutputSchedulerState 	**outputSchedule=scheduleState->outputSched;
	int i;
	
	outputSchedule[output] = (OutputSchedulerState *) malloc( sizeof( OutputSchedulerState ) );
	outputSchedule[output]->request = (int *) malloc( sizeof(int)*aSwitch->numInputs);
	
	outputSchedule[output]->RR_grants = (int *) malloc( sizeof(int)*aSwitch->numInputs );
	
	outputSchedule[output]->GBVOQ_grants = (int *)malloc( sizeof(int)*aSwitch->numInputs ) ;
	
#ifdef	DTC
	outputSchedule[output]->slackness_grants = (int *)malloc( sizeof(int)*aSwitch->numInputs ) ;
	outputSchedule[output]->slackness_inputs = (int *)malloc( sizeof(int)*aSwitch->numInputs ) ;
#endif
	
	for(i=0; i<aSwitch->numInputs; i++ )
	{
		outputSchedule[output]->RR_grants[i] = i;
		
		outputSchedule[output]->GBVOQ_grants[i] = NONE ;
		if( DTC )		
		{
			outputSchedule[output]->slackness_grants[i] = 123456789 ;
			outputSchedule[output]->slackness_inputs[i] = NONE ;
		}
	}
	outputSchedule[output]->last_accepted_grant=0;
	outputSchedule[output]->unsynch = 0 ;
	
	return(outputSchedule[output]);
}

#ifdef USE_ALL_FUNCS

static void
printState(aSwitch, aState)
Switch *aSwitch;
State *aState;
{
	int input, output, numFabricOutputs;
	
	numFabricOutputs = aSwitch->numOutputs*aSwitch->fabric.Xbar_numOutputLines;
	
	printf("Switch %2d:", aSwitch->switchNumber);
	for(output=0; output<numFabricOutputs; output++)
		printf("%3d ", output);
	printf("   Output: ");
	for(output=0; output<numFabricOutputs; output++)
		printf("%2d ", output);
	printf("Input: ");
	for(input=0; input<aSwitch->numInputs; input++)
		printf("%2d ", input);
	printf("\n          ");
	/* Connection matrix */
	for(output=0; output<numFabricOutputs; output++)
    {
		if( aState->connection[output] == NONE )
			printf("  X ");
		else
			printf("%3d ", aState->connection[output] );
    }
	printf("           ");
	/* Output schedulers */
	for(output=0; output<numFabricOutputs; output++)
		printf("%2d ", aState->outputGrantPointer[output]);
	printf("       ");
	/* Input schedulers */
	for(input=0; input<aSwitch->numInputs; input++)
		printf("%2d ", aState->inputAcceptPointer[input]);
	printf("\n");
	printf("-----------------------------------------------------------\n");
	
}
#endif


int l_search(int array[], int n, int key)
{
	int i ;
	
	for(i=0; i<n; i++)
		if(array[i] == key)
			return i ;
	
	return NONE ;
}

void dtc_sort(int slackness[], int slackness_length, int port[], int port_length)
{
    int from, to, save_slackness, save_port ;
	
    for(from=0; from<slackness_length; from++)
        for(to=from+1; to<slackness_length; to++)
            if(slackness[from] > slackness[to])
            {
                save_slackness = slackness[from] ;
                save_port = port[from] ;
                slackness[from] = slackness[to] ;
                port[from] = port[to] ;
                slackness[to] = save_slackness ;
                port[to] = save_port ;
            }
}

void setupGrants(aSwitch, output)	/* CREATE INPUT SCHEDULE FOR THIS OUTPUT. */
Switch	*aSwitch ;
int		output ;
{
	SchedulerState	*scheduleState = (SchedulerState *)aSwitch->scheduler.schedulingState ;
	OutputSchedulerState *outputSchedule = scheduleState->outputSched[output] ;
	
	struct Element			*inElement ;
	Cell					*inCell ;
	int						inPort, inIndex, g ;
	
	/* Store the alternative inputs provided by the OPL in order to resolve the conflicts appeared in switch s scheduling table. */
	
	/* Initialize GBVOQ_grants array. */
	for(g=0; g<aSwitch->numOutputs; g++)
		outputSchedule->GBVOQ_grants[g] = NONE ;
	
	EACH_ELEMENT(aSwitch->outputBuffer[output]->outputPriorityList, inElement)
	{
		inCell = (Cell *)inElement->Object ;
		inPort = inCell->commonStats.inputPort ;
		if( DTC )	inCell->commonStats.slackness = priorObjectsInList(aSwitch->outputBuffer[output]->outputPriorityList, inCell)-priorObjectsInList(aSwitch->inputBuffer[inPort]->inputPriorityList, inCell) ;
		/* Search for the input that appears in the OPL, but not in GBVOQ_grants. */
		if(l_search(outputSchedule->GBVOQ_grants, aSwitch->numOutputs, inPort) == NONE)
		{
			/* Store the inputs in the order they appear in the OPL. */
			inIndex = l_search(outputSchedule->GBVOQ_grants, aSwitch->numOutputs, NONE) ;
			if(inIndex != NONE)
			{
				outputSchedule->GBVOQ_grants[inIndex] = inPort ;
				if( DTC )
				{
					outputSchedule->slackness_grants[inIndex] = inCell->commonStats.slackness ;
					outputSchedule->slackness_inputs[inIndex] = inCell->commonStats.inputPort ;
				}
			}
			else	break ;
		}
	}
	
	if( DTC )	
	{
		if( debug_algorithm )
		{
			fprintf(stdout, "BEFORE : GRANTS : ") ;
			for(g=0; g<aSwitch->numOutputs; g++)
				fprintf(stdout, "%d ", outputSchedule->slackness_inputs[g]) ;
			fprintf(stdout, "\n") ;
			fprintf(stdout, "BEFORE : SLACKNESS : ") ;
			for(g=0; g<aSwitch->numOutputs; g++)
				fprintf(stdout, "%d ", outputSchedule->slackness_grants[g]) ;
			fprintf(stdout, "\n") ;
		}
		
		dtc_sort(outputSchedule->slackness_grants, aSwitch->numOutputs, outputSchedule->slackness_inputs, aSwitch->numOutputs) ;
		for(g=0; g<aSwitch->numOutputs; g++)
			outputSchedule->GBVOQ_grants[g] = outputSchedule->slackness_inputs[g] ;
		
		if( debug_algorithm )
		{
			fprintf(stdout, "AFTER : GRANTS : ") ;
			for(g=0; g<aSwitch->numOutputs; g++)
				fprintf(stdout, "%d ", outputSchedule->slackness_inputs[g]) ;
			fprintf(stdout, "\n") ;
			fprintf(stdout, "AFTER : SLACKNESS : ") ;
			for(g=0; g<aSwitch->numOutputs; g++)
				fprintf(stdout, "%d ", outputSchedule->slackness_grants[g]) ;
			fprintf(stdout, "\n") ;
		}
	}
	
	if( debug_algorithm )
	{
		/*fprintf(stdout, "---> OUTPUT_PORT =%d | OPL : ", output) ;
		 if( aSwitch->outputBuffer[output]->outputPriorityList->head != NULL )
		 {
		 EACH_ELEMENT(aSwitch->outputBuffer[output]->outputPriorityList, inElement)
		 {
		 inCell = (Cell *)inElement->Object ;
		 fprintf(stdout, "%d ", inCell->commonStats.inputPort) ;
		 }
		 fprintf(stdout, "\n") ;
		 }
		 else	fprintf(stdout, "\n") ;*/
		fprintf(stdout, "---> OUTPUT_PORT =%d | GBVOQ_grants : ", output) ;
		for(g=0; g<aSwitch->numOutputs; g++)
			fprintf(stdout, "%d ", scheduleState->outputSched[output]->GBVOQ_grants[g]) ;
		fprintf(stdout, "\n") ;
	}
	
}

void setupAccepts(aSwitch, input)	/* CREATE OUTPUT SCHEDULE FOR THIS INPUT. */
Switch	*aSwitch ;
int		input ;
{
	SchedulerState	*scheduleState = (SchedulerState *)aSwitch->scheduler.schedulingState ;
	InputSchedulerState *inputSchedule = scheduleState->inputSched[input] ;
	
	struct Element	*outElement ;
	Cell			*outCell ;
	int				outPort, outIndex, a ;
	
	/* Store the alternative inputs provided by the OPL in order to resolve the conflicts appeared in switch s scheduling table. */
	
	/* Initialize GBVOQ_accepts array. */
	for(a=0; a<aSwitch->numInputs; a++)
		inputSchedule->GBVOQ_accepts[a] = NONE ;
	
	EACH_ELEMENT(aSwitch->inputBuffer[input]->inputPriorityList, outElement)
	{
		outCell = (Cell *)outElement->Object ;
		outPort = outCell->vci ;
		if( DTC )	outCell->commonStats.slackness = priorObjectsInList(aSwitch->outputBuffer[outPort]->outputPriorityList, outCell)-priorObjectsInList(aSwitch->inputBuffer[input]->inputPriorityList, outCell) ;
		/* Search for the output that appears in the IPL, but not in GBVOQ_accepts.*/
		if(l_search(inputSchedule->GBVOQ_accepts, aSwitch->numInputs, outPort) == NONE)
		{
			// Store the outputs in the order they appear in the IPL.
			outIndex = l_search(inputSchedule->GBVOQ_accepts, aSwitch->numInputs, NONE) ;
			if(outIndex != NONE)
			{
				inputSchedule->GBVOQ_accepts[outIndex] = outPort ;
				if( DTC )
				{
					inputSchedule->slackness_accepts[outIndex] = outCell->commonStats.slackness ;
					inputSchedule->slackness_outputs[outIndex] = outCell->vci ;
				}
			}
			else	break ;
		}
	}
	
	if( DTC )	
	{
		if( debug_algorithm )
		{
			fprintf(stdout, "BEFORE : ACCEPTS : ") ;
			for(a=0; a<aSwitch->numInputs; a++)
				fprintf(stdout, "%d ", inputSchedule->slackness_outputs[a]) ;
			fprintf(stdout, "\n") ;
			fprintf(stdout, "BEFORE : SLACKNESS : ") ;
			for(a=0; a<aSwitch->numInputs; a++)
				fprintf(stdout, "%d ", inputSchedule->slackness_accepts[a]) ;
			fprintf(stdout, "\n") ;
		}
		
		dtc_sort(inputSchedule->slackness_accepts, aSwitch->numInputs, inputSchedule->slackness_outputs, aSwitch->numInputs) ;
		for(a=0; a<aSwitch->numInputs; a++)
			inputSchedule->GBVOQ_accepts[a] = inputSchedule->slackness_outputs[a] ;
		
		if( debug_algorithm )
		{
			fprintf(stdout, "BEFORE : ACCEPTS : ") ;
			for(a=0; a<aSwitch->numInputs; a++)
				fprintf(stdout, "%d ", inputSchedule->slackness_outputs[a]) ;
			fprintf(stdout, "\n") ;
			fprintf(stdout, "BEFORE : SLACKNESS : ") ;
			for(a=0; a<aSwitch->numInputs; a++)
				fprintf(stdout, "%d ", inputSchedule->slackness_accepts[a]) ;
			fprintf(stdout, "\n") ;
		}
	}
	
	if( debug_algorithm )
	{
		/*fprintf(stdout, "---> INPUT_PORT =%d | IPL : ", input) ;
		 if( aSwitch->inputBuffer[input]->inputPriorityList->head != NULL )
		 {
		 EACH_ELEMENT(aSwitch->inputBuffer[input]->inputPriorityList, outElement)
		 {
		 outCell = (Cell *)outElement->Object ;
		 fprintf(stdout, "%d ", outCell->vci) ;
		 }
		 fprintf(stdout, "\n") ;
		 }
		 else	fprintf(stdout, "\n") ;*/
		fprintf(stdout, "---> INPUT_PORT =%d | GBVOQ_accepts : ", input) ;
		for(a=0; a<aSwitch->numInputs; a++)
			fprintf(stdout, "%d ", inputSchedule->GBVOQ_accepts[a]) ;
		fprintf(stdout, "\n") ;
	}
	
}


void displayGrantTokens(aSwitch, input)
Switch	*aSwitch ;
int		input ;
{
	int i ;
	
	fprintf(stdout, "$$$ PBC $$$ : inputBuffer[%d]->grantQueue | ", input) ;
	for(i=0; i<aSwitch->numOutputs; i++)
		fprintf(stdout, "%2d ", aSwitch->inputBuffer[input]->grantQueue[i]) ;
	fprintf(stdout, "\n") ;
}

void displayCreditTokens(aSwitch)
Switch	*aSwitch ;
{
	int i ;
	
	fprintf(stdout, "$$$ PBC $$$ : creditQueue | ") ;
	for(i=0; i<aSwitch->numOutputs; i++)
		fprintf(stdout, "[%d:]%d ", i, aSwitch->outputBuffer[i]->creditQueue) ;
	fprintf(stdout, "\n") ;
}

int calculateGrantTokens(aSwitch, input)
Switch	*aSwitch ;
int		input ;
{
	int i, tokens ;
	
	tokens = 0 ;
	for(i=0; i<aSwitch->numOutputs; i++)
		if( aSwitch->inputBuffer[input]->grantQueue[i] != NONE )
			tokens ++ ;
	
	return tokens ;
}

void calculateCreditTokens(aSwitch, output)
Switch	*aSwitch ;
int		output ;
{
	int iB, tokens ;
	
	tokens = 0 ;
	for(iB=0; iB<aSwitch->numInternalBuffers; iB++)
		if( aSwitch->outputBuffer[output]->fifo[iB]->head == NULL )
			tokens ++ ;
	
	aSwitch->outputBuffer[output]->creditQueue = tokens ;
}

int scheduleCreditTokens(aSwitch)
Switch	*aSwitch ;
{
	int i, tokens ;
	
	tokens = 0 ;
	for(i=0; i<aSwitch->numOutputs; i++)
		tokens += aSwitch->outputBuffer[i]->creditQueue ;
	
	return tokens ;
}

int remainingGrantTokens(aSwitch, input)
Switch	*aSwitch ;
int		input ;
{
	int i, tokens ;
	
	tokens = 0 ;
	for(i=0; i<aSwitch->numOutputs; i++)
		if( aSwitch->inputBuffer[input]->grantQueue[i] == 1 )
			tokens ++ ;
	
	return tokens ;
}

void resetGrantTokens(aSwitch, input)
Switch	*aSwitch ;
int		input ;
{
	int i ;
	
	for(i=0; i<aSwitch->numOutputs; i++)
		aSwitch->inputBuffer[input]->grantQueue[i] = NONE ;
}

void resetRemainingGrantTokens(aSwitch, input, voq)
Switch	*aSwitch ;
int		input ;
int		voq ;
{
	int i ;
	
	for(i=0; i<aSwitch->numOutputs; i++)
	{
		if( i == voq )	continue ;
		if( aSwitch->inputBuffer[input]->grantQueue[i] == NONE )	continue ;
		aSwitch->inputBuffer[input]->grantQueue[i] = 0 ;
	}
}

void displayGrants(aSwitch)
Switch	*aSwitch ;
{
	int i ;
	
	fprintf(stdout, "$$$ PBC $$$ : gbvoqQueue | ") ;
	for(i=0; i<aSwitch->numOutputs; i++)
		fprintf(stdout, "[%d:]%d ", i, aSwitch->outputBuffer[i]->gbvoqQueue) ;
	fprintf(stdout, "\n") ;
}

void calculateGrants(aSwitch, output)
Switch	*aSwitch ;
int		output ;
{
	int i, tokens ;
	SchedulerState	*scheduleState = (SchedulerState *)aSwitch->scheduler.schedulingState ;
	OutputSchedulerState *outputSchedule = scheduleState->outputSched[output] ;
	
	tokens = 0 ;
	for(i=0; i<aSwitch->numInputs; i++)
		if( outputSchedule->GBVOQ_grants[i] != NONE )
			tokens ++ ;
	
	aSwitch->outputBuffer[output]->gbvoqQueue = tokens ;
}

int scheduleGrants(aSwitch)
Switch	*aSwitch ;
{
	int i, tokens ;
	
	tokens = 0 ;
	for(i=0; i<aSwitch->numOutputs; i++)
		tokens += aSwitch->outputBuffer[i]->gbvoqQueue ;
	
	return tokens ;
}